###############################################################################
#
#   Cyclotomics
"""
    Cyclotomic(n, coeffs::AbstractVector)
Element of `n`-th cyclotomic field with coefficients stored as `coeffs`.

To access the internals of a cyclotomic use API functions:
 * `conductor` - the conductor of a cyclotomic, i.e. the `n` used currently for
 storage. This might not be the minimal embeding field of a cyclotomic.
 * `getindex`/`setindex!` - use `α[i]` to access the coefficient at `i`-th power
 of a cyclotomic (in a circular fashion)
 * `values`/`exponents` - paired iterators over _non zero_ coefficients/exponents
 corresponding to _non-zero_ coefficients
 * `normalform!` - bring a cyclotomic to its unique representation as given by
 Zumbroich basis (also available in non-modifying form).

Iteration over non-zero coefficients in `Cyclotomic` is provided by `iterate`
which produces pairs `(exp, coeff)` of exponent and corresponding coefficient.
"""
struct Cyclotomic{T,A<:AbstractVector{T}} <: Number
    n::Int
    coeffs::A
end

Cyclotomic(v::V) where {V<:AbstractVector} =
    Cyclotomic{eltype(v),V}(length(v), v)
Cyclotomic{T}(α::Cyclotomic) where {T} =
    Cyclotomic(conductor(α), convert.(T, coeffs(α)))
Cyclotomic{T,V}(α::Cyclotomic) where {T,V} =
    Cyclotomic{T,V}(conductor(α), convert.(T, coeffs(α)))

Cyclotomic{T,V}(a::R) where {T,V,R<:Real} = Cyclotomic{T,V}(1, T[a])

"""
    E(n[, i=1])
Return the `i`-th power of `n`-th root of unity with sparse vector as storage.
"""
function E(n, i = 1)
    k = totient(n)
    i = (0 <= i < n ? i : mod(i, n))
    coeffs = sparsevec([i + 1], [1], n)
    sizehint!(coeffs.nzind, k)
    sizehint!(coeffs.nzval, k)
    return Cyclotomic(n, coeffs)
end

####
#   Low-level interface

"""
    conductor(α::Cyclotomic)
Return the conductor, i.e. the degree of cyclotomic field `α` belongs to.
"""
conductor(α::Cyclotomic) = α.n

"""
    coeffs(α::Cyclotomic)
Return the coefficients of `α` as stored.
"""
coeffs(α::Cyclotomic) = α.coeffs

function _to_index(α::Cyclotomic, idx::Integer)
    # return mod(idx, conductor(α)) + 1
    0 <= idx < conductor(α) && return idx + 1
    conductor(α) <= idx && return (idx % conductor(α)) + 1
    return (idx % conductor(α)) + conductor(α) + 1
end

Base.@propagate_inbounds function Base.getindex(α::Cyclotomic, exp::Integer)
    return α.coeffs[_to_index(α, exp)]
end

Base.getindex(α::Cyclotomic, itr) = [α[i] for i in itr]

Base.@propagate_inbounds function Base.setindex!(
    α::Cyclotomic,
    val,
    exp::Integer,
)
    α.coeffs[_to_index(α, exp)] = val
    return val
end

# Base.@propagate_inbounds function Base.setindex!(α::Cyclotomic, val, itr)
#     for exp in itr
#         α[exp] = val
#     end
#     return itr
# end

# general definitions for iteration
function Base.iterate(α::Cyclotomic, state = 0)
    idx = findnext(!iszero, coeffs(α), state + 1)
    idx === nothing && return nothing
    return (idx - 1, coeffs(α)[idx]), idx
end

Base.IteratorSize(::Type{<:Cyclotomic}) = Base.HasLength()
Base.length(α::Cyclotomic) = count(!iszero, coeffs(α))
Base.eltype(::Type{<:Cyclotomic{T}}) where {T} = Tuple{Int,T}

"""
    exponents(α::Cyclotomic)
Return an iterator over non-zero exponents of `α`, beginning at `0`-th one.
Matched iterator over coefficients is provided by @ref(values).
"""
exponents(α::Cyclotomic) = (first(i) for i in α)

"""
    values(α::Cyclotomic)
Return an iterator over non-zero coefficients of `α`, beginning at `0`-th one.
Matched iterator over exponents is provided by @ref(exponents).
"""
Base.values(α::Cyclotomic) = (last(i) for i in α)
Base.valtype(::Cyclotomic{T}) where {T} = T

Base.similar(α::Cyclotomic, T::Type = valtype(α)) = similar(α, T, conductor(α))
Base.similar(α::Cyclotomic, m::Integer) = similar(α, valtype(α), m)
Base.similar(α::Cyclotomic, T::Type, n::Integer) =
    Cyclotomic(similar(coeffs(α), T, n))

"""
    dense(α::Cyclotomic)
Return a copy of `α` with coefficients stored in dense `Vector`.
"""
dense(α::Cyclotomic{T}) where {T} =
    Cyclotomic{T,Vector{T}}(conductor(α), coeffs(α))

"""
    sparse(α::Cyclotomic)
Return a copy of `α` with coefficients stored in `SparseVector`.
"""
SparseArrays.sparse(α::Cyclotomic) = Cyclotomic(sparse(coeffs(α)))

function Base.float(::Type{T}, α::Cyclotomic) where {T}
    β = reduced_embedding(α)
    isreal(β) && return real(Complex{T}(β))
    throw(InexactError(:float, AbstractFloat, α))
end

Base.float(α::Cyclotomic) = float(Float64, α)

for f in (:Int, :Float64)
    q = QuoteNode(f)
    @eval begin
        function Base.$f(α::Cyclotomic)
            β = reduced_embedding(α)
            isreal(β) && return $f(real(ComplexF64(β)))
            throw(InexactError($q, $f, α))
        end
    end
end

function Base.Complex{T}(α::Cyclotomic) where {T<:AbstractFloat}
    z = zero(Complex{T})
    rα = reduced_embedding(α)
    n = conductor(rα)
    for (e, c) in rα
        γ = 2 * T(π) * T(e) / n
        z += c * (cos(γ) + im * sin(γ))
    end
    return z
end

Base.abs2(α::Cyclotomic) = α * conj(α)
